﻿using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;

public enum AttackType {
    None,
    Dash,
    Slam,
    Arm,
    Leg,
}

public struct AttackInfo {
    public Bounds bounds;
    public Vector2 direction;
    public AttackType type;
}

public class Tom : MonoBehaviour {

    private float m_hAcceleration = 40.0f;
    private float m_hDeceleration = 90.0f;
    private float m_hSpeed = 10.0f;

    private float m_inAirFactor = 0.25f;
    private float m_jumpSpeed = 25.0f;

    private float m_slamSpeed = 30.0f;
    private float m_slamPauseDuration = 0.15f;
    private float m_slamCooldownDuration = 1.5f;

    private float m_gravity = 9.81f;
    private float m_gravityFactor = 7.0f;

    private float m_dashSpeed = 30.0f;
    private float m_dashDuration = 0.25f;
    private float m_dashStartMaximumSpeed = 10.0f;

    private float m_armHitDuration = 0.1f;
    private float m_legHitDuration = 0.2f;
    private float m_facingDirection = 1.0f;
    private Bounds m_standingHitBox = new Bounds( );
    private Bounds m_dashingHitBox = new Bounds( );

    private Alarm m_inputPause = new Alarm( );
    private Alarm m_dashAlarm = new Alarm( );
    private Alarm m_hitAlarmArm = new Alarm( );
    private Alarm m_hitAlarmLeg = new Alarm( );
    private Alarm m_recovery = new Alarm( );
    private Alarm m_slamCooldown = new Alarm( );

    private int m_health = 10;
    private float m_recoveryDuration = 1.0f;
    private int m_flash = 0;

    private bool m_grounded = false;
    private Vector2 m_currentSpeed = Vector2.zero;
    private AttackInfo m_attackInfo = new AttackInfo( );

    public GameObject m_arm = null;
    public GameObject m_leg = null;
    public GameObject m_healthPrefab = null;
    public GameObject m_slamBar = null;
    public Spawner m_spawner = null;

    private GameObject[ ] m_healthSquares = null;

    private Transform m_root = null;
    public Animator m_animator = null;
    public Transform m_visual = null;
    public SpriteRenderer m_sprite = null;

    public AudioClip m_attack = null;
    public AudioClip m_dash = null;
    public AudioClip m_slam = null;
    public AudioSource m_audioSource = null;
    public Text m_endText = null;

    void Start( ) {

        m_root = transform.parent;
        m_arm.SetActive( false );
        m_leg.SetActive( false );

        m_healthSquares = new GameObject[ m_health ];

        float screenSize = Camera.main.orthographicSize;
        float cameraOffset = Camera.main.transform.position.y;
        float ratio = 16.0f / 9.0f;

        for ( int healthIndex = 0; healthIndex < m_health; healthIndex++ ) {
            m_healthSquares[ healthIndex ] = Instantiate( m_healthPrefab ) as GameObject;
            m_healthSquares[ healthIndex ].transform.parent = Camera.main.transform;
            m_healthSquares[ healthIndex ].transform.position = new Vector3( -screenSize * ratio + 1.0f + healthIndex * 0.5f, screenSize - 0.5f + cameraOffset );
        }

        m_slamBar = Instantiate( m_slamBar ) as GameObject;
        m_slamBar.transform.parent = Camera.main.transform;
        m_slamBar.transform.position = new Vector3( screenSize * ratio - 1.0f, screenSize - 0.5f + cameraOffset );

        m_slamCooldown.start( 0.01f );
    }

    void Update( ) {

        m_standingHitBox = new Bounds( transform.position - transform.localPosition + new Vector3( 0.0f, 1.0f ), new Vector3( 0.5f, 2.0f ) );
        m_dashingHitBox = new Bounds( transform.position - transform.localPosition + new Vector3( 0.0f, 0.25f ), new Vector3( 0.5f, 0.5f ) );

        if ( Input.GetButtonDown( "Restart" ) ) {
            Application.LoadLevel( 0 );
        }

        m_inputPause.step( );
        m_dashAlarm.step( );
        m_hitAlarmArm.step( );
        m_recovery.step( );
        m_slamCooldown.step( );

        m_slamBar.transform.localScale = new Vector3( Mathf.Clamp01( m_slamCooldown.m_age / m_slamCooldown.m_duration ) * 10.0f, 1.0f );

        if ( m_recovery.isRunning( ) ) {

            m_flash = ( m_flash + 1 ) % 10;

            if ( m_flash > 4 ) {
                m_sprite.enabled = false;
            } else {
                m_sprite.enabled = true;
            }
        } else {
            m_sprite.enabled = true;
        }

        bool inputPause = m_inputPause.isRunning( );

        if ( !inputPause ) {
            m_attackInfo.type = AttackType.None;
        }

        float hInput = 0.0f;

        hInput = Input.GetAxisRaw( "horizontal" );

        if ( hInput < -0.25f ) {
            hInput = -1.0f;
        } else if ( hInput > 0.25f ) {
            hInput = 1.0f;
        } else {
            hInput = 0.0f;
        }

        if ( ( inputPause && !m_hitAlarmLeg.isRunning( ) ) || m_health <= 0 ) {
            hInput = 0.0f;
        }

        if ( ( !inputPause || m_hitAlarmLeg.isRunning( ) ) && Input.GetButton( "Left" ) && m_health > 0 ) {
            hInput -= 1.0f;
        }

        if ( ( !inputPause || m_hitAlarmLeg.isRunning( ) ) && Input.GetButton( "Right" ) && m_health > 0 ) {
            hInput += 1.0f;
        }

        float hSign = Mathf.Sign( m_currentSpeed.x );
        float airFactor = 1.0f;

        if ( hSign == hInput ) {
            m_facingDirection = hInput;
        }

        if ( !m_grounded ) {
            airFactor = m_inAirFactor;
        }

        if ( hInput != 0.0f ) {

            if ( hSign == 0.0f || hSign == hInput ) {
                m_currentSpeed.x += m_hAcceleration * hInput * airFactor * Time.deltaTime;
            } else {
                m_currentSpeed.x += m_hDeceleration * hInput * airFactor * Time.deltaTime;
            }

        } else {

            m_currentSpeed.x += m_hDeceleration * airFactor * -hSign * Time.deltaTime;

            if ( hSign != Mathf.Sign( m_currentSpeed.x ) ) {
                m_currentSpeed.x = 0.0f;
            }
        }

        if ( !m_dashAlarm.isRunning( ) && Mathf.Abs( m_currentSpeed.x ) > m_hSpeed ) {
            m_currentSpeed.x = Mathf.Sign( m_currentSpeed.x ) * m_hSpeed;
        }

        if ( !inputPause && Input.GetButtonDown( "Jump" ) && m_grounded && m_health > 0 ) {
            m_currentSpeed.y = m_jumpSpeed;
            m_grounded = false;
            m_animator.Play( "jump" );
        }

        if ( !inputPause && Input.GetButtonDown( "Dash" ) ) {

            if ( !m_grounded ) {

                if ( !m_slamCooldown.isRunning( ) && m_currentSpeed.y <= m_dashStartMaximumSpeed ) {
                    m_currentSpeed.y = -m_slamSpeed;
                    m_currentSpeed.x = 0.0f;
                    m_inputPause.start( m_slamPauseDuration );
                    m_attackInfo.type = AttackType.Slam;
                    m_slamCooldown.start( m_slamCooldownDuration );
                    m_animator.Play( "slam" );
                    m_audioSource.clip = m_slam;
                    m_audioSource.volume = 1.0f;
                    m_audioSource.Play( );
                }

            } else {

                if ( !m_slamCooldown.isRunning( ) && m_currentSpeed.x != 0.0f ) {
                    m_dashAlarm.start( m_dashDuration );
                    m_inputPause.start( m_dashDuration );
                    m_currentSpeed.x = Mathf.Sign( m_currentSpeed.x ) * m_dashSpeed;
                    m_attackInfo.type = AttackType.Dash;
                    m_slamCooldown.start( m_slamCooldownDuration );
                    m_animator.Play( "dash" );
                    m_audioSource.clip = m_dash;
                    m_audioSource.volume = 1.0f;
                    m_audioSource.Play( );
                }
            }
        }

        m_currentSpeed.y += -m_gravity * m_gravityFactor * Time.deltaTime;
        m_root.position += new Vector3( m_currentSpeed.x, m_currentSpeed.y ) * Time.deltaTime;

        if ( m_root.position.y < 0 ) {
            m_root.position = new Vector3( m_root.position.x, 0.0f );
            m_currentSpeed.y = 0.0f;
            m_grounded = true;

            if ( m_hitAlarmLeg.isRunning( ) ) {
                m_hitAlarmLeg.stop( );
                m_inputPause.stop( );
            }

            if ( !m_inputPause.isRunning( ) && !isAnimatorPlaying( "run" ) && !isAnimatorPlaying( "slam" ) && !isAnimatorPlaying( "dash" ) ) {
                m_animator.Play( "idle" );
            }
        }

        if ( m_root.position.x < -8.0f ) {
            m_root.position = new Vector3( -8.0f, m_root.position.y );
        }

        if ( m_root.position.x > 8.0f ) {
            m_root.position = new Vector3( 8.0f, m_root.position.y );
        }

        if ( !inputPause && Input.GetButtonDown( "Attack" ) && m_health > 0 ) {

            if ( m_grounded && !m_hitAlarmArm.isRunning( ) ) {

                m_arm.SetActive( true );
                Vector3 armPosition = m_arm.transform.localPosition;
                m_hitAlarmArm.start( m_armHitDuration );
                m_arm.transform.localPosition = new Vector3( 0.875f * m_facingDirection, armPosition.y, armPosition.z );
                m_inputPause.start( m_armHitDuration );
                m_attackInfo.type = AttackType.Arm;
                m_animator.Play( "attack_arm" );
                m_audioSource.clip = m_attack;
                m_audioSource.volume = 0.5f;
                m_audioSource.Play( );

            } else if ( !m_grounded && !m_hitAlarmLeg.isRunning( ) ) {

                m_leg.SetActive( true );
                Vector3 legPosition = m_leg.transform.localPosition;
                m_hitAlarmLeg.start( m_legHitDuration );
                m_leg.transform.localPosition = new Vector3( 1.0f * m_facingDirection, legPosition.y, legPosition.z );
                m_inputPause.start( m_legHitDuration );
                m_attackInfo.type = AttackType.Leg;
                m_animator.Play( "attack_leg" );
                m_audioSource.clip = m_attack;
                m_audioSource.volume = 0.5f;
                m_audioSource.Play( );
            }
        }

        if ( !m_hitAlarmArm.isRunning( ) ) {
            m_arm.SetActive( false );
        }

        if ( !m_hitAlarmLeg.isRunning( ) ) {
            m_leg.SetActive( false );
        }

        if ( m_health > 0 && !m_recovery.isRunning( ) && !m_inputPause.isRunning( ) ) {

            Bounds self = m_standingHitBox;

            if ( m_hitAlarmLeg.isRunning( ) || isAnimatorPlaying( "dash" ) ) {
                self = m_dashingHitBox;
            }
            
            List<Blob> blobs = m_spawner.getBlobs( );

            for ( int blobIndex = 0; blobIndex < blobs.Count; blobIndex++ ) {

                AttackInfo info = blobs[ blobIndex ].getHitBox( );

                if ( m_health > 0 && info.type != AttackType.None && info.bounds.Intersects( self ) ) {
                    m_health--;
                    m_healthSquares[ m_health ].SetActive( false );
                    m_recovery.start( m_recoveryDuration );
                    m_currentSpeed.x += ( transform.position.x - info.bounds.center.x ) * 100.0f;
                    m_currentSpeed.y += 10.0f;
                    m_grounded = false;
                }
            }
        }

        if ( !isAnimatorPlaying( "attack_arm" ) && !isAnimatorPlaying( "dash" ) && !isAnimatorPlaying( "slam" ) ) {

            if ( m_grounded && Mathf.Abs( m_currentSpeed.x ) < 0.1f ) {
                m_animator.Play( "idle" );
            } else if ( m_grounded ) {
                m_animator.Play( "run" );
            }
        }

        m_visual.localScale = new Vector3( Mathf.Abs( m_visual.localScale.x ) * m_facingDirection * -1.0f, m_visual.localScale.y );

        if ( m_health <= 0 ) {
            m_endText.enabled = true;
        } else {
            m_endText.enabled = false;
        }
    }

    public bool isAnimatorPlaying( string animation ) {
        bool result = m_animator.GetCurrentAnimatorStateInfo( 0 ).IsName( animation );
        return result;
    }

    public AttackInfo getAttackInfo( ) {

        switch ( m_attackInfo.type ) {

            case AttackType.Slam: {
                m_attackInfo.bounds = new Bounds( transform.position, transform.localScale );
                m_attackInfo.direction = new Vector2( 0.0f, -1.0f );
            } break;

            case AttackType.Dash: {
                m_attackInfo.bounds = new Bounds( transform.position, transform.localScale );
                m_attackInfo.direction = new Vector2( Mathf.Sign( m_currentSpeed.x ), 0.0f );
            } break;

            case AttackType.Arm: {
                m_attackInfo.bounds = new Bounds( m_arm.transform.position, m_arm.transform.localScale );
                m_attackInfo.direction = new Vector2( m_facingDirection, 0.0f );
            } break;

            case AttackType.Leg: {
                m_attackInfo.bounds = new Bounds( m_leg.transform.position, m_leg.transform.localScale );
                m_attackInfo.direction = new Vector2( m_facingDirection, 0.0f );
            } break;
        }

        return m_attackInfo;
    }

    void OnDrawGizmos( ) {

        Gizmos.color = Color.blue;

        Bounds b = m_standingHitBox;

        if ( m_hitAlarmLeg.isRunning( ) ) {
            b = m_dashingHitBox;
        }

        Gizmos.DrawWireCube( b.center, b.size );

        AttackInfo info = getAttackInfo( );

        if ( info.type != AttackType.None ) {
            Gizmos.color = Color.yellow;
            Gizmos.DrawWireCube( info.bounds.center, info.bounds.size );
        }
    }
}
